# This code is copyrighted by Stephan Richter and
# Thomas Foerster in September, 2001.
#
# written by Stephan Richter
# Last updated: October 4, 2001

# *** this file implements the semantics of the ZOQL ***
#
# example:
# def example_method( list, context):
#     method body
#
# list - list of parsed elements of the rule,
# e.g. ['CREATE', 'PRODUCT', 'Test'...
# nonterminals come out as a list, rather than relying on the nesting
# level you should write another semantic function to retrieve the item
# itself (or use one of the helpers --> item, item1 ...)
# context - global parse context (variables etc.), if any ;-)

import string, sys

import Operators
from ZOQLExceptions import *

# global method binding dictionary
# key is a semantic method name
# value is list with syntax rules bound to that method 
bindings = {}


def Nothing(list, context):
    return list

# I list 'unassigned' rules here, meaning they have to be assigned
# to more tough functions working with the objects referenced
bindings['Nothing'] = ["assignment1", "assignment2", "assignment3",
                       "assignment4","assignment5", "assignment6",
                       "parenth_form"]


### Helper functions
def get_identifier_name(elem):
    if type(elem).__name__ == 'instance':
        return elem.attr
    else:
        return elem


def item(list, context):
    return list[0]

bindings['item'] = ["product_type1", "product_type2", "opt_recursive1",
                    "selection1", "selection2",
                    "opt_prop_declaration1", "prop_declaration_list1",
                    "opt_order_dir1", "opt_order_dir2",
                    "atom1", "atom2", "atom3", "enclosure1",
                    "enclosure2", "enclosure3", "enclosure4",
                    "key_datum_list1", "literal1", "literal2",
                    "literal3", "literal4", "literal5", "primary1", "primary2",
                    "primary3", "opt_argument_list1", "reference1",
                    "reference2", "power1", "u_expr1", "m_expr1",
                    "a_expr1", "parameter1", "defparameter2",
                    "opt_parameter_list1", "opt_def_param_list1",
                    "expression_list1", "not_test1", "and_test1",
                    "or_test1", "expression2", "or_expr1", "and_expr1",
                    "xor_expr1", "shift_expr1", "opt_index_declaration1"]


def item1(list, context):
    return list[1]

bindings['item1'] = ["opt_inherit1", "opt_propertysheet_declaration1", "opt_offset1",
                     "opt_assignment1", "opt_option1", "opt_from1", "opt_in1", "opt_default1",
                     "list_display", "dict_display", "parameter2", "index_list"]


def item2(list, context):
    return list[2]

bindings['item2'] = ["opt_dastident1", "opt_orderby1"]


def item3(list, context):
    return list[3]


def empty(list, context):
    return None

bindings['empty'] = ["opt_propertysheet_declaration0",
                     "opt_assignment0", "opt_restrict0",
                     "opt_from0", "opt_recursive0", "opt_orderby0",
                     "opt_default0", "opt_argument_list0",
                     "opt_dastident0", "opt_comp_expr0"]


def empty_list(list, context):
    return []

bindings['empty_list'] = ["opt_inherit0", 'opt_param_list_append0',
                          "opt_parameter_list0", "opt_def_param_list0", 
                          "opt_prop_declaration0", "def_param_list0"]#, "index_declaration_list1"]


def one_element_list(list, context):
    return [list[0]]

bindings['one_element_list'] = ["propertysheet_list1", "index_declaration_list1",
                                "order_list0"]


def insert_element(list, context):
    """ Prepend the first element to the list in the third element
    Assumes list to be: [element, ',', element_list]
    """
    
    if type(list[2]) == type([]):
        result = list[2]
    else:
        result = [list[2]]
        
    result.insert(0, list[0])
    return result

bindings['insert_element'] = ["assignment_list2", "propertysheet_list2",
                              "prop_declaration_list2", "reference_list2", "order_list1", 
                              "def_param_list1", "expression_list2", "index_declaration_list2"]


def append_element(list, context):
    """ Append the third list element to the first element list
    Assumes list to be: [element_list, ',', element]
    """
    if type(list[2]) == type([]):
        result = list[2]
    else:
        result = [list[2]]
    
    result.append(list[2])
    return result


def toString(list, context):
    result = str(list[1])
    return result

bindings['toString'] = ["string_conversion"]


def joinlist(list, context):
    list = map(get_identifier_name, list)
    result = string.join(list,'')
    return result

bindings['joinlist'] = ["attribute_ref"]


def create_dictionary(list, context):
    """ Convert a [key, ':', value] list to dictionary """ 
    result = {}
    result[list[0]] = list[2]
    return result

bindings['create_dictionary'] = ["key_datum"]



# *** Starting Arithmetic Operators ***

def add(list, context):
    return Operators.Add(list[0], list[2])

bindings['add'] = ["key_datum_list2", "a_expr2"]
    

def minus(list, context):
    return Operators.Subtract(list[0], list[2])

bindings['minus'] = ["a_expr3"]


def mul(list, context):
    return Operators.Multiply(list[0], list[2])

bindings['mul'] = ["m_expr2"]


def power(list, context):
    return Operators.Power(list[0], list[2])

bindings['power'] = ["power2"]


def div(list, context):
    return Operators.Divide(list[0], list[2])

bindings['div'] = ["m_expr3"]


def mod(list, context):
    return Operators.Modular(list[0], list[2])

bindings['mod'] = ["m_expr4"]



# *** Starting Boolean Operators ***

def not_func(list, context):
    return (1, Operators.Not(None, list[1]))

bindings['not_func'] = ["not_test2"]
    
    
def and_func(list, context):
    return (1, Operators.And(list[0], list[2]))

bindings['and_func'] = ["and_test2"]


def or_func(list, context):
    return (1, Operators.Or(list[0], list[2]))

bindings['or_func'] = ["or_test2"]


def xor_func(list, context):
    return (1, Operators.Xor(list[0], list[2]))

bindings['xor_func'] = ["xor_expr2"]



# *** Starting PropertySheet Handlers ***

def get_propertysheet(list, context):
    """ Turn propertysheet into dictionary """
    result = {}
    result['name'] = get_identifier_name(list[0])
    result['props'] = list[2]
    return result

bindings['get_propertysheet'] = ["propertysheet"]


def get_property(list, context):
    """ Turn a property into a dictionary """
    result = {}
    result['name'] = get_identifier_name(list[0])
    result['type'] = get_identifier_name(list[2])
    result['default'] = get_identifier_name(list[3])
    return result

bindings['get_property'] = ["prop_declaration"]


# *** Starting Index Handlers ***

def get_index_declaration(list, context):
    ''' '''
    index = {}
    index['name'] = get_identifier_name(list[0])
    index['type'] = get_identifier_name(list[2])
    return index

bindings['get_index_declaration'] = ["index_declaration"]


# *** Starting Limit Handlers ***

def get_limit(list, context):
    ''' '''
    limit = {}
    limit['size'] = list[1]
    limit['start'] = list[2]
    if not limit['start']:
        limit['start'] = 0
    return limit

bindings['get_limit'] = ["opt_limit1"]


# *** Starting Limit Handlers ***

def get_order_item(list, context):
    ''' '''
    order = {}
    order['property'] = get_identifier_name(list[0])
    order['direction'] = list[1]
    if not order['direction']:
        order['direction'] = 'ASC'
    return order

bindings['get_order_item'] = ["order_item"]


# *** Starting General Handlers ***

def sublist(list,context):
    """ Handling of sublists """
    if list[1] is not None:
        result = [list[0]].append(list[1])
    elif list[2] == ',':
        result = [list[0]]
    else:
        result = list[0]

    return result

bindings['sublist'] = ["sublist"]


def param_list_append(list, context):
    result = [list[1]].append(list[2])
    return result

bindings['param_list_append'] = ["opt_param_list_append1"]


def itemlist(list, context):
    if list[1] is not None:
        result = [list[0]]
    else:
        result = list[0]

    return result

bindings['itemlist'] = ["last_element3"]


def last_element1(list, context):
    result = {}
    result['*'] = list[1]
    result['**'] = list[2]

    return result

bindings['last_element1'] = ["last_element1"]


def last_element2(list, context):
    result={}
    result['**'] = list[1]

    return result

bindings['last_element2'] = ["last_element2"]


# *** Start of Comparison Operators ***

def comparison(list, context):

    lhs = list[0]
    if list[1] is not None:
        ## really a comparison
        
        val, rhs = list[1][1] # bool value and lhs expr of rhs comparison
        symbol = list[1][0]

        if symbol == '==':
            return (1, Operators.Equal(lhs, rhs))
        if symbol == '!=' or symbol == '<>':
            return (1, Operators.NotEqual(lhs, rhs))
        if symbol == '<=':
            return (1, Operators.LessThanEqual(lhs, rhs))
        if symbol == '>=':
            return (1, Operators.GreaterThanEqual(lhs, rhs))
        if symbol == '<':
            return (1, Operators.LessThan(lhs, rhs))
        if symbol == '>':
            return (1, Operators.GreaterThan(lhs, rhs))
        else:
            raise UnknownComparisonOperator(symbol)
    else:
        return (1, lhs)

bindings['comparison'] = ["comparison"]


# *** Start of Declarative (Assignment) Operators ***

def set_equal(list, context):
    return Operators.SetEqual(list[0], list[2])

bindings['set_equal'] = ["assignment1"]


def plus_equal(list, context):
    return Operators.PlusEqual(list[0], list[2])

bindings['plus_equal'] = ["assignment2"]


def minus_equal(list, context):
    return Operators.MinusEqual(list[0], list[2])

bindings['minus_equal'] = ["assignment3"]


def mul_equal(list, context):
    return Operators.MulEqual(list[0], list[2])

bindings['mul_equal'] = ["assignment4"]


def div_equal(list, context):
    return Operators.DivEqual(list[0], list[2])

bindings['div_equal'] = ["assignment5"]


def mod_equal(list, context):
    return Operators.ModEqual(list[0], list[2])

bindings['mod_equal'] = ["assignment6"]



# *** Start of Miscelleneous Handlers ***

def expression(list, context):
    return list[0][1]

bindings['expression'] = ["expression1"]


def itemaslist(list, context):
    return [list[0]]

bindings['itemaslist'] = ["reference_list1"]


def item1aslist(list, context):
    return [list[1]]

bindings['item1aslist'] = ["opt_restrict1"]


# *** Meta Type List Handler ***

def get_meta_type_list(raw_meta_types):
    raw_meta_types = map(get_identifier_name, raw_meta_types)
    meta_types = []
    for meta_type in raw_meta_types:
        if type(meta_type) in (type(()), type([])):
            meta_types += list(meta_type)
        else:
            meta_types.append(meta_type)
    return meta_types


# *** Beginning of Statement Handlers ***

def create_product(list, context):
    """Create a new product"""

    product_id = get_identifier_name(list[2])
    product_type = list[4]
    
    if context.has_key('base_object') and context.has_key('executer'):
        return context['executer'].do_create_product(context['base_object'], product_id, product_type)
    else:
        raise BadContextError()


bindings['create_product'] = ["statement1"]



def create_class(list, context):
    """ Create a class in a product """

    python_products = []

    # collect data from list
    class_id = get_identifier_name(list[2])
    prod_id = get_identifier_name(list[4])
    attributes = list[5]
    options = list[6]
    propertysheets = list[7]
    raw_bases = list[8]

    # Now, because of the ZClasses' sort of dumb way of referencing base classes,
    # we got to do some cleanup
    bases = []
    for base in raw_bases:
        base = string.replace(base, '/', '.')
        base = string.split(base, '.')
        base = string.join([string.join(base[:-1], '.'), base[-1]], '/') 
        bases.append(base)
            
    if context.has_key('base_object') and context.has_key('executer'):
        if prod_id in python_products:
            pass
        else:
            context['executer'].do_create_zclass(context['base_object'], prod_id, class_id,
                                                 attributes, options, propertysheets, bases)
    else:
        raise BadContextError()


bindings['create_class'] = ["statement2"]


def create_index(list, context):
    """ Create a ZOQL index using ZCatalog """

    name = get_identifier_name(list[2])
    selected_meta_types = map(get_identifier_name, list[4])
    index_list = list[5]
    conditions = list[6]
    if conditions:
        conditions = conditions[0]

    if list[7]: recursive = 1
    else: recursive = 0

    if context.has_key('base_object') and context.has_key('executer'):
        return context['executer'].do_create_index(context['base_object'], name, selected_meta_types,
                                                   index_list, conditions, recursive)
    else:
        raise BadContextError()


bindings['create_index'] = ["statement3"]


def insert_statement(list, context):
    """ Perform an INSERT """

    meta_type = get_identifier_name(list[1])
    container_path = get_identifier_name(list[3])
    assignments = list[4]

    if context.has_key('base_object') and context.has_key('executer'):
        return context['executer'].do_insert(context['base_object'], meta_type, container_path, assignments)
    else:
        raise BadContextError()


bindings['insert_statement'] = ["statement4"]



def update_statement(list, context):
    """ Perform an UPDATE """

    selected_meta_types = get_meta_type_list(list[1])
    searched_containers = map(get_identifier_name, list[2])
    assignments = list[4]
    conditions = list[5]
    if conditions:
        conditions = conditions[0]

    if list[6]: recursive = 1
    else: recursive = 0

    if context.has_key('base_object') and context.has_key('executer'):
        return context['executer'].do_update(context['base_object'], selected_meta_types, searched_containers,
                                             assignments, conditions, recursive)
    else:
        raise BadContextError()


bindings['update_statement'] = ["statement5"]



def delete_statement(list, context):

    selected_meta_types = get_meta_type_list(list[1])
    searched_containers = map(get_identifier_name, list[2])
    conditions = list[3]
    if conditions:
        conditions = conditions[0]

    if list[4]: recursive = 1
    else: recursive = 0

    if context.has_key('base_object') and context.has_key('executer'):
        return context['executer'].do_delete(context['base_object'], selected_meta_types,
                                             searched_containers, conditions, recursive)
    else:
        raise BadContextError()


bindings['delete_statement'] = ["statement6"]



def select_statement(list, context):

    selected_meta_types = get_meta_type_list(list[1])
    searched_containers = map(get_identifier_name, list[2])
    conditions = list[3]
    if conditions:
        conditions = conditions[0]
    orderby = list[4]
    limit = list[5]
    if list[6]: recursive = 1
    else: recursive = 0

    if context.has_key('base_object') and context.has_key('executer'):
        result = context['executer'].do_select(context['base_object'], selected_meta_types,
                                               searched_containers, conditions, orderby, limit, recursive)
        context['result'].append(result)
    else:
        raise BadContextError()
        

bindings['select_statement'] = ["statement7"]



# *** Binding of rules to semantic functions ***

def BindRules(Grammar):
    """ Single place for method bindings """

    ## there are two ways to bind rules to functions
    ## either use the bindings dictionary
    ##     bindings[funcname] = [list_of_rulenames]
    
    for func in bindings.keys():
        for rule in bindings[func]:
            Grammar.Bind(rule, globals()[func])
            
    ## additional bindings
    ## example: Grammar.Bind("Rulename", example_method)
    for i in range(1,11):
        Grammar.Bind("comparison_operator"+str(i), item)
